from __future__ import annotations

import os
import subprocess
from pathlib import Path

import click

SHELLS = ["bash", "zsh", "fish"]
shell = os.environ.get("SHELL")

if shell is not None:
    shell = Path(shell).name
    if shell not in SHELLS:
        shell = None


@click.group(
    help="Generate shell completion",
)
def completion() -> None:
    pass


@completion.command(help="Generate shell completion code")
@click.option(
    "-s",
    "--shell",
    type=click.Choice(SHELLS),
    default=shell,
    show_default=True,
    required=shell is None,
    help="Shell type to generate for",
)
def generate(shell: str) -> None:
    """
    \b
    Generate shell completion code for the current, or specified (-s)hell.
    You will need to 'source' this code to enable shell completion.
    You can source it directly (performs slower) by running:
        \033[3;33meval "$(chik complete generate)"\033[0m
    or you can save the output to a file:
        \033[3;33mchik complete generate > ~/.chik-complete-bash\033[0m
    and source that file with:
        \033[3;33m. ~/.chik-complete-bash\033[0m
    """
    # Could consider calling directly in the future.
    # https://github.com/pallets/click/blob/ef11be6e49e19a055fe7e5a89f0f1f4062c68dba/src/click/shell_completion.py#L17
    subprocess.run(["chik"], check=True, env={**os.environ, "_CHIK_COMPLETE": f"{shell}_source"})



from __future__ import annotations

import json
import os
import shlex
import urllib
import uuid
import webbrowser
from collections.abc import Sequence
from pathlib import Path
from typing import Callable, ClassVar, Literal, Optional, Union, overload

import anyio
import click
import yaml

from chik.cmds.cmd_classes import chik_command, option


class UnexpectedFormError(Exception):
    pass


Oses = Union[Literal["linux"], Literal["macos-arm"], Literal["macos-intel"], Literal["windows"]]
Method = Union[Literal["GET"], Literal["POST"]]
Per = Union[Literal["directory"], Literal["file"]]

all_oses: Sequence[Oses] = ("linux", "macos-arm", "macos-intel", "windows")


def report(*args: str) -> None:
    print("    ====", *args)


async def get_gh_token() -> str:
    command = ["gh", "auth", "token"]
    report(f"running command: {shlex.join(command)}")
    process = await anyio.run_process(command=command, check=False, stderr=None)

    if process.returncode != 0:
        raise click.ClickException("failed to get gh cli personal access token")

    return process.stdout.decode("utf-8").strip()


@overload
async def run_gh_api(method: Method, args: list[str], error: str) -> None: ...
@overload
async def run_gh_api(method: Method, args: list[str], error: str, capture_stdout: Literal[False]) -> None: ...
@overload
async def run_gh_api(method: Method, args: list[str], error: str, capture_stdout: Literal[True]) -> str: ...


async def run_gh_api(method: Method, args: list[str], error: str, capture_stdout: bool = False) -> Optional[str]:
    command = [
        "gh",
        "api",
        f"--method={method}",
        "-H=Accept: application/vnd.github+json",
        "-H=X-GitHub-Api-Version: 2022-11-28",
        *args,
    ]
    report(f"running command: {shlex.join(command)}")

    if capture_stdout:
        process = await anyio.run_process(command=command, check=False, stderr=None)
    else:
        process = await anyio.run_process(command=command, check=False, stderr=None, stdout=None)

    if process.returncode != 0:
        raise click.ClickException(error)

    if capture_stdout:
        return process.stdout.decode("utf-8")

    return None


def input_arg(name: str, value: object, cond: bool = True) -> list[str]:
    if not cond:
        return []

    assert value is not None

    if isinstance(value, os.PathLike):
        value = os.fspath(value)
    dumped = yaml.safe_dump(value).partition("\n")[0]
    return [f"-f=inputs[{name}]={dumped}"]


@click.group("gh", help="For working with GitHub")
def gh_group() -> None:
    pass


@chik_command(
    group=gh_group,
    name="test",
    short_help="launch a test run in CI from HEAD or existing remote ref",
    help="""Allows easy triggering and viewing of test workflow runs in CI including
    configuration of parameters.  If a ref is specified then it must exist on the
    remote and a run will be launched for it.  If ref is not specified then the local
    HEAD will be pushed to a temporary remote branch and a run will be launched for
    that.  There is no need to push the local commit first.  The temporary remote
    branch will automatically be deleted in most cases.

    After launching the workflow run GitHub will be queried for the run and the URL
    will be opened in the default browser.
    """,
)
class TestCMD:
    workflow_id: ClassVar[str] = "test.yml"
    owner: str = option("-o", "--owner", help="Owner of the repo", type=str, default="Chik-Network")
    repository: str = option("-r", "--repository", help="Repository name", type=str, default="chik-blockchain")
    ref: Optional[str] = option(
        "-f",
        "--ref",
        help="Branch or tag name (commit SHA not supported), if not specified will push HEAD to a temporary branch",
        type=str,
        default=None,
    )
    per: Per = option("-p", "--per", help="Per", type=click.Choice(["directory", "file"]), default="directory")
    only: Optional[Path] = option(
        "-o", "--only", help="Only run this item, a file or directory depending on --per", type=Path
    )
    duplicates: int = option("-d", "--duplicates", help="Number of duplicates", type=int, default=1)
    oses: Sequence[Oses] = option(
        "--os",
        help="Operating systems to run on",
        type=click.Choice(all_oses),
        multiple=True,
        default=all_oses,
    )
    full_python_matrix: bool = option(
        "--full-python-matrix/--default-python-matrix", help="Run on all Python versions", default=False
    )
    remote: str = option("-r", "--remote", help="Name of git remote", type=str, default="origin")
    open_browser: bool = option("--open-browser/--no-open-browser", help="Open browser", default=True)

    async def run(self) -> None:
        await self.check_only()

        username = await self.get_username()

        if self.ref is not None:
            await self.trigger_workflow(self.ref)
            query = "+".join(
                [
                    "event=workflow_dispatch",
                    f"branch={self.ref}",
                    f"actor={username}",
                ]
            )
            run_url = f"https://github.com/Chik-Network/chik-blockchain/actions/workflows/test.yml?query={urllib.parse.quote(query)}"
            report(f"waiting a few seconds to load: {run_url}")
            await anyio.sleep(10)
        else:
            process = await anyio.run_process(command=["git", "rev-parse", "HEAD"], check=True, stderr=None)
            if process.returncode != 0:
                raise click.ClickException("Failed to get current commit SHA")

            commit_sha = process.stdout.decode("utf-8").strip()

            temp_branch_name = f"tmp/{username}/{commit_sha}/{uuid.uuid4()}"

            process = await anyio.run_process(
                command=["git", "push", self.remote, f"HEAD:{temp_branch_name}"], check=False, stdout=None, stderr=None
            )
            if process.returncode != 0:
                raise click.ClickException("Failed to push temporary branch")

            try:
                await self.trigger_workflow(temp_branch_name)
                for _ in range(10):
                    await anyio.sleep(1)

                    try:
                        report("looking for run")
                        html_url, api_url = await self.find_run(temp_branch_name)
                        report(f"run found at: {html_url}")
                        report(f"run found at: {api_url}")
                    except UnexpectedFormError:
                        report("run not found")
                        continue

                    break
                else:
                    raise click.ClickException("Failed to find run url")
            finally:
                report(f"deleting temporary branch: {temp_branch_name}")
                process = await anyio.run_process(
                    command=["git", "push", self.remote, "-d", temp_branch_name], check=False, stdout=None, stderr=None
                )
                if process.returncode != 0:
                    raise click.ClickException("Failed to dispatch workflow")
                report(f"temporary branch deleted: {temp_branch_name}")

        report(f"run html url: {html_url}")
        report(f"run api url: {api_url}")
        if self.open_browser:
            webbrowser.open(html_url)

    async def check_only(self) -> None:
        if self.only is not None:
            import chik._tests

            test_path = Path(chik._tests.__file__).parent
            effective_path = test_path.joinpath(self.only)
            checks: dict[Per, Callable[[], bool]] = {"directory": effective_path.is_dir, "file": effective_path.is_file}
            check = checks[self.per]
            if not check():
                if effective_path.exists():
                    explanation = "wrong type"
                else:
                    explanation = "does not exist"
                message = f"expected requested --only to be a {self.per}, {explanation} at: {effective_path.as_posix()}"
                raise click.ClickException(message)

    async def trigger_workflow(self, ref: str) -> None:
        # https://docs.github.com/en/rest/actions/workflows?apiVersion=2022-11-28#create-a-workflow-dispatch-event
        await run_gh_api(
            method="POST",
            args=[
                f"/repos/{self.owner}/{self.repository}/actions/workflows/{self.workflow_id}/dispatches",
                f"-f=ref={ref}",
                *input_arg("per", self.per),
                *input_arg("only", self.only, self.only is not None),
                *input_arg("duplicates", self.duplicates),
                *(arg for os_name in all_oses for arg in input_arg(f"run-{os_name}", os_name in self.oses)),
                *input_arg("full-python-matrix", self.full_python_matrix),
            ],
            error="Failed to dispatch workflow",
        )
        report(f"workflow triggered on branch: {ref}")

    async def find_run(self, ref: str) -> tuple[str, str]:
        # https://docs.github.com/en/rest/actions/workflow-runs?apiVersion=2022-11-28#list-workflow-runs-for-a-workflow
        stdout = await run_gh_api(
            method="GET",
            args=[
                f"-f=branch={ref}",
                f"/repos/{self.owner}/{self.repository}/actions/workflows/{self.workflow_id}/runs",
            ],
            error="Failed to query workflow runs",
            capture_stdout=True,
        )

        response = json.loads(stdout)
        runs = response["workflow_runs"]
        try:
            [run] = runs
        except ValueError:
            raise UnexpectedFormError(f"expected 1 run, got: {len(runs)}")

        html_url = run["html_url"]
        assert isinstance(html_url, str), f"expected html url to be a string, got: {html_url!r}"

        api_url = run["url"]
        assert isinstance(api_url, str), f"expected url to be a string, got: {api_url!r}"

        return html_url, api_url

    async def get_username(self) -> str:
        # https://docs.github.com/en/rest/users/users?apiVersion=2022-11-28#get-the-authenticated-user
        stdout = await run_gh_api(
            method="GET",
            args=["/user"],
            error="Failed to get username",
            capture_stdout=True,
        )

        response = json.loads(stdout)
        username = response["login"]
        assert isinstance(username, str), f"expected username to be a string, got: {username!r}"
        return username



"""
Pay to delegated puzzle

In this puzzle program, the solution must be a signed delegated puzzle, along with
its (unsigned) solution. The delegated puzzle is executed, passing in the solution.
This obviously could be done recursively, arbitrarily deep (as long as the maximum
cost is not exceeded).

If you want to specify the conditions directly (thus terminating the potential recursion),
you can use p2_conditions.

This roughly corresponds to bitcoin's graftroot.
"""

from __future__ import annotations

from chik.types.blockchain_format.program import Program
from chik.wallet.puzzles import p2_conditions
from chik.wallet.puzzles.load_klvm import load_klvm_maybe_recompile

MOD = load_klvm_maybe_recompile("p2_delegated_puzzle.clsp")


def puzzle_for_pk(public_key: bytes) -> Program:
    return MOD.curry(public_key)


def solution_for_conditions(conditions) -> Program:
    delegated_puzzle = p2_conditions.puzzle_for_conditions(conditions)
    return solution_for_delegated_puzzle(delegated_puzzle, Program.to(0))


def solution_for_delegated_puzzle(delegated_puzzle: Program, delegated_solution: Program) -> Program:
    return delegated_puzzle.to([delegated_puzzle, delegated_solution])



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from chik.protocols.wallet_protocol import CoinState
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint16, uint64
from chik.util.streamable import Streamable, streamable
from chik.wallet.lineage_proof import LineageProof


@streamable
@dataclass(frozen=True)
class DIDInfo(Streamable):
    origin_coin: Optional[Coin]  # Coin ID of this coin is our DID
    backup_ids: list[bytes32]
    num_of_backup_ids_needed: uint64
    parent_info: list[tuple[bytes32, Optional[LineageProof]]]  # {coin.name(): LineageProof}
    current_inner: Optional[Program]  # represents a Program as bytes
    temp_coin: Optional[Coin]  # partially recovered wallet uses these to hold info
    temp_puzhash: Optional[bytes32]
    temp_pubkey: Optional[bytes]
    sent_recovery_transaction: bool
    metadata: str  # JSON of the user defined metadata


@streamable
@dataclass(frozen=True)
class DIDCoinData(Streamable):
    p2_puzzle: Program
    recovery_list_hash: bytes32
    num_verification: uint16
    singleton_struct: Program
    metadata: Program
    inner_puzzle: Optional[Program]
    coin_state: CoinState



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, ClassVar, Optional, cast

from chik.introducer.introducer import Introducer
from chik.protocols.introducer_protocol import RequestPeersIntroducer, RespondPeersIntroducer
from chik.protocols.protocol_message_types import ProtocolMessageTypes
from chik.rpc.rpc_server import StateChangedProtocol
from chik.server.api_protocol import ApiMetadata
from chik.server.outbound_message import Message, make_msg
from chik.server.ws_connection import WSChikConnection
from chik.types.peer_info import TimestampedPeerInfo
from chik.util.ints import uint64


class IntroducerAPI:
    if TYPE_CHECKING:
        from chik.server.api_protocol import ApiProtocol

        _protocol_check: ClassVar[ApiProtocol] = cast("IntroducerAPI", None)

    log: logging.Logger
    introducer: Introducer
    metadata: ClassVar[ApiMetadata] = ApiMetadata()

    def __init__(self, introducer) -> None:
        self.log = logging.getLogger(__name__)
        self.introducer = introducer

    def ready(self) -> bool:
        return True

    def _set_state_changed_callback(self, callback: StateChangedProtocol) -> None:
        pass

    @metadata.request(peer_required=True)
    async def request_peers_introducer(
        self,
        request: RequestPeersIntroducer,
        peer: WSChikConnection,
    ) -> Optional[Message]:
        max_peers = self.introducer.max_peers_to_send
        if self.introducer.server is None or self.introducer.server.introducer_peers is None:
            return None
        rawpeers = self.introducer.server.introducer_peers.get_peers(
            max_peers * 5, True, self.introducer.recent_peer_threshold
        )

        peers = []
        for r_peer in rawpeers:
            if r_peer.vetted <= 0:
                continue

            if r_peer.host == peer.peer_info.host and r_peer.port == peer.peer_server_port:
                continue
            peer_without_timestamp = TimestampedPeerInfo(
                r_peer.host,
                r_peer.port,
                uint64(0),
            )
            peers.append(peer_without_timestamp)

            if len(peers) >= max_peers:
                break

        self.introducer.log.info(f"Sending vetted {peers}")

        msg = make_msg(ProtocolMessageTypes.respond_peers_introducer, RespondPeersIntroducer(peers))
        return msg



from __future__ import annotations

from collections.abc import Iterable

from chik_rs import AugSchemeMPL, G1Element

from chik._tests.klvm.coin_store import CoinStore, CoinTimestamp
from chik._tests.core.make_block_generator import int_to_public_key
from chik._tests.util.key_tool import KeyTool
from chik.consensus.default_constants import DEFAULT_CONSTANTS
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import make_spend
from chik.types.spend_bundle import SpendBundle
from chik.util.hash import std_hash
from chik.util.ints import uint32, uint64
from chik.wallet.puzzles import (
    p2_conditions,
    p2_delegated_conditions,
    p2_delegated_puzzle,
    p2_delegated_puzzle_or_hidden_puzzle,
    p2_m_of_n_delegate_direct,
    p2_puzzle_hash,
)
from chik.wallet.puzzles.puzzle_utils import make_create_coin_condition

T1 = CoinTimestamp(1, uint32(10000000))
T2 = CoinTimestamp(5, uint32(10003000))

MAX_BLOCK_COST_KLVM = 10**18


def secret_exponent_for_index(index: int) -> int:
    blob = index.to_bytes(32, "big")
    hashed_blob = AugSchemeMPL.key_gen(std_hash(b"foo" + blob))
    r = int.from_bytes(hashed_blob, "big")
    return r


def public_key_for_index(index: int, key_lookup: KeyTool) -> bytes:
    secret_exponent = secret_exponent_for_index(index)
    key_lookup.add_secret_exponents([secret_exponent])
    return bytes(int_to_public_key(secret_exponent))


def throwaway_puzzle_hash(index: int, key_lookup: KeyTool) -> bytes32:
    return p2_delegated_puzzle.puzzle_for_pk(public_key_for_index(index, key_lookup)).get_tree_hash()


def do_test_spend(
    puzzle_reveal: Program,
    solution: Program,
    payments: Iterable[tuple[bytes32, int]],
    key_lookup: KeyTool,
    farm_time: CoinTimestamp = T1,
    spend_time: CoinTimestamp = T2,
) -> SpendBundle:
    """
    This method will farm a coin paid to the hash of `puzzle_reveal`, then try to spend it
    with `solution`, and verify that the created coins correspond to `payments`.

    The `key_lookup` is used to create a signed version of the `SpendBundle`
    """

    coin_db = CoinStore(DEFAULT_CONSTANTS)

    puzzle_hash = puzzle_reveal.get_tree_hash()

    # farm it
    coin = coin_db.farm_coin(puzzle_hash, farm_time)

    # spend it
    coin_spend = make_spend(coin, puzzle_reveal, solution)

    # sign the solution
    signature = key_lookup.signature_for_solution(coin_spend, DEFAULT_CONSTANTS.AGG_SIG_ME_ADDITIONAL_DATA)
    spend_bundle = SpendBundle([coin_spend], signature)
    coin_db.update_coin_store_for_spend_bundle(spend_bundle, spend_time, MAX_BLOCK_COST_KLVM)

    # ensure all outputs are there
    for puzzle_hash, amount in payments:
        for coin in coin_db.coins_for_puzzle_hash(puzzle_hash):
            if coin.amount == amount:
                break
        else:
            assert 0

    return spend_bundle


def default_payments_and_conditions(
    initial_index: int, key_lookup: KeyTool
) -> tuple[list[tuple[bytes32, int]], Program]:
    # the coin we get from coin_db.farm_coin only has amount 1024, so we can
    # only make small payments to avoid failing with MINTING_COIN
    payments = [
        (throwaway_puzzle_hash(initial_index + 1, key_lookup), initial_index * 10),
        (throwaway_puzzle_hash(initial_index + 2, key_lookup), (initial_index + 1) * 10),
    ]
    conditions = Program.to([make_create_coin_condition(ph, uint64(amount), []) for ph, amount in payments])
    return payments, conditions


def test_p2_conditions():
    key_lookup = KeyTool()
    payments, conditions = default_payments_and_conditions(1, key_lookup)

    puzzle = p2_conditions.puzzle_for_conditions(conditions)
    solution = p2_conditions.solution_for_conditions(conditions)

    do_test_spend(puzzle, solution, payments, key_lookup)


def test_p2_delegated_conditions():
    key_lookup = KeyTool()
    payments, conditions = default_payments_and_conditions(1, key_lookup)

    pk = public_key_for_index(1, key_lookup)

    puzzle = p2_delegated_conditions.puzzle_for_pk(pk)
    solution = p2_delegated_conditions.solution_for_conditions(conditions)

    do_test_spend(puzzle, solution, payments, key_lookup)


def test_p2_delegated_puzzle_simple():
    key_lookup = KeyTool()
    payments, conditions = default_payments_and_conditions(1, key_lookup)

    pk = public_key_for_index(1, key_lookup)

    puzzle = p2_delegated_puzzle.puzzle_for_pk(pk)
    solution = p2_delegated_puzzle.solution_for_conditions(conditions)

    do_test_spend(puzzle, solution, payments, key_lookup)


def test_p2_delegated_puzzle_graftroot():
    key_lookup = KeyTool()
    payments, conditions = default_payments_and_conditions(1, key_lookup)

    delegated_puzzle = p2_delegated_conditions.puzzle_for_pk(public_key_for_index(8, key_lookup))
    delegated_solution = p2_delegated_conditions.solution_for_conditions(conditions)

    puzzle_program = p2_delegated_puzzle.puzzle_for_pk(public_key_for_index(1, key_lookup))
    solution = p2_delegated_puzzle.solution_for_delegated_puzzle(delegated_puzzle, delegated_solution)

    do_test_spend(puzzle_program, solution, payments, key_lookup)


def test_p2_puzzle_hash():
    key_lookup = KeyTool()
    payments, conditions = default_payments_and_conditions(1, key_lookup)

    inner_puzzle = p2_delegated_conditions.puzzle_for_pk(public_key_for_index(4, key_lookup))
    inner_solution = p2_delegated_conditions.solution_for_conditions(conditions)
    inner_puzzle_hash = inner_puzzle.get_tree_hash()

    puzzle_program = p2_puzzle_hash.puzzle_for_inner_puzzle_hash(inner_puzzle_hash)
    assert puzzle_program == p2_puzzle_hash.puzzle_for_inner_puzzle(inner_puzzle)
    solution = p2_puzzle_hash.solution_for_inner_puzzle_and_inner_solution(inner_puzzle, inner_solution)

    do_test_spend(puzzle_program, solution, payments, key_lookup)


def test_p2_m_of_n_delegated_puzzle():
    key_lookup = KeyTool()
    payments, conditions = default_payments_and_conditions(1, key_lookup)

    pks = [public_key_for_index(_, key_lookup) for _ in range(1, 6)]
    M = 3

    delegated_puzzle = p2_conditions.puzzle_for_conditions(conditions)
    delegated_solution = []

    puzzle_program = p2_m_of_n_delegate_direct.puzzle_for_m_of_public_key_list(M, pks)
    selectors = [1, [], [], 1, 1]
    solution = p2_m_of_n_delegate_direct.solution_for_delegated_puzzle(
        M, selectors, delegated_puzzle, delegated_solution
    )

    do_test_spend(puzzle_program, solution, payments, key_lookup)


def test_p2_delegated_puzzle_or_hidden_puzzle_with_hidden_puzzle():
    key_lookup = KeyTool()
    payments, conditions = default_payments_and_conditions(1, key_lookup)

    hidden_puzzle = p2_conditions.puzzle_for_conditions(conditions)
    hidden_public_key = public_key_for_index(10, key_lookup)

    puzzle = p2_delegated_puzzle_or_hidden_puzzle.puzzle_for_public_key_and_hidden_puzzle(
        G1Element.from_bytes_unchecked(hidden_public_key), hidden_puzzle
    )
    solution = p2_delegated_puzzle_or_hidden_puzzle.solution_for_hidden_puzzle(
        G1Element.from_bytes_unchecked(hidden_public_key), hidden_puzzle, Program.to(0)
    )

    do_test_spend(puzzle, solution, payments, key_lookup)


def do_test_spend_p2_delegated_puzzle_or_hidden_puzzle_with_delegated_puzzle(hidden_pub_key_index):
    key_lookup = KeyTool()
    _payments, conditions = default_payments_and_conditions(1, key_lookup)

    hidden_puzzle = p2_conditions.puzzle_for_conditions(conditions)
    hidden_public_key = public_key_for_index(hidden_pub_key_index, key_lookup)
    hidden_pub_key_point = G1Element.from_bytes(hidden_public_key)

    puzzle = p2_delegated_puzzle_or_hidden_puzzle.puzzle_for_public_key_and_hidden_puzzle(
        hidden_pub_key_point, hidden_puzzle
    )
    payable_payments, payable_conditions = default_payments_and_conditions(5, key_lookup)

    delegated_puzzle = p2_conditions.puzzle_for_conditions(payable_conditions)
    delegated_solution = []

    synthetic_public_key = p2_delegated_puzzle_or_hidden_puzzle.calculate_synthetic_public_key(
        G1Element.from_bytes(hidden_public_key), hidden_puzzle.get_tree_hash()
    )

    solution = p2_delegated_puzzle_or_hidden_puzzle.solution_for_delegated_puzzle(delegated_puzzle, delegated_solution)

    hidden_puzzle_hash = hidden_puzzle.get_tree_hash()
    synthetic_offset = p2_delegated_puzzle_or_hidden_puzzle.calculate_synthetic_offset(
        hidden_pub_key_point, hidden_puzzle_hash
    )

    assert synthetic_public_key == int_to_public_key(synthetic_offset) + hidden_pub_key_point

    secret_exponent = key_lookup.dict[G1Element.from_bytes(hidden_public_key)]
    assert int_to_public_key(secret_exponent) == hidden_pub_key_point

    synthetic_secret_exponent = secret_exponent + synthetic_offset
    key_lookup.add_secret_exponents([synthetic_secret_exponent])

    do_test_spend(puzzle, solution, payable_payments, key_lookup)


def test_p2_delegated_puzzle_or_hidden_puzzle_with_delegated_puzzle():
    for hidden_pub_key_index in range(1, 10):
        do_test_spend_p2_delegated_puzzle_or_hidden_puzzle_with_delegated_puzzle(hidden_pub_key_index)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from chik.types.spend_bundle_conditions import SpendBundleConditions
from chik.util.ints import uint16
from chik.util.streamable import Streamable, streamable


@streamable
@dataclass(frozen=True)
class NPCResult(Streamable):
    error: Optional[uint16]
    conds: Optional[SpendBundleConditions]



from __future__ import annotations

from math import ceil

import pytest

from chik.util.paginator import InvalidPageSizeError, InvalidPageSizeLimit, PageOutOfBoundsError, Paginator


@pytest.mark.parametrize(
    "source, page_size, page_size_limit",
    [([], 1, 1), ([1], 1, 2), ([1, 2], 2, 2), ([], 10, 100), ([1, 2, 10], 1000, 1000)],
)
def test_constructor_valid_inputs(source: list[int], page_size: int, page_size_limit: int) -> None:
    paginator: Paginator = Paginator.create(source, page_size, page_size_limit)
    assert paginator.page_size() == page_size
    assert paginator.page_count() == 1
    assert paginator.get_page(0) == source


@pytest.mark.parametrize(
    "page_size, page_size_limit, exception",
    [
        (5, -1, InvalidPageSizeLimit),
        (5, 0, InvalidPageSizeLimit),
        (2, 1, InvalidPageSizeError),
        (100, 1, InvalidPageSizeError),
        (1001, 1000, InvalidPageSizeError),
    ],
)
def test_constructor_invalid_inputs(page_size: int, page_size_limit: int, exception: type[Exception]) -> None:
    with pytest.raises(exception):
        Paginator.create([], page_size, page_size_limit)


def test_page_count() -> None:
    for page_size in range(1, 10):
        for i in range(0, 10):
            assert Paginator.create(range(0, i), page_size).page_count() == max(1, ceil(i / page_size))


@pytest.mark.parametrize(
    "length, page_size, page, expected_data",
    [
        (17, 5, 0, [0, 1, 2, 3, 4]),
        (17, 5, 1, [5, 6, 7, 8, 9]),
        (17, 5, 2, [10, 11, 12, 13, 14]),
        (17, 5, 3, [15, 16]),
        (3, 4, 0, [0, 1, 2]),
        (3, 3, 0, [0, 1, 2]),
        (3, 2, 0, [0, 1]),
        (3, 2, 1, [2]),
        (3, 1, 0, [0]),
        (3, 1, 1, [1]),
        (3, 1, 2, [2]),
        (2, 2, 0, [0, 1]),
        (2, 1, 0, [0]),
        (2, 1, 1, [1]),
        (1, 2, 0, [0]),
        (0, 2, 0, []),
        (0, 10, 0, []),
    ],
)
def test_get_page_valid(length: int, page: int, page_size: int, expected_data: list[int]) -> None:
    assert Paginator.create(list(range(0, length)), page_size).get_page(page) == expected_data


@pytest.mark.parametrize("page", [-1000, -10, -1, 5, 10, 1000])
def test_get_page_invalid(page: int) -> None:
    with pytest.raises(PageOutOfBoundsError):
        Paginator.create(range(0, 17), 5).get_page(page)



from __future__ import annotations

import json
from dataclasses import dataclass
from io import StringIO
from typing import Optional, cast

import pytest
from chik_rs import G1Element

# TODO: update after resolution in https://github.com/pytest-dev/pytest/issues/7469
from pytest_mock import MockerFixture

from chik._tests.cmds.cmd_test_utils import TestWalletRpcClient
from chik._tests.conftest import ConsensusMode
from chik._tests.environments.wallet import WalletStateTransition, WalletTestFramework
from chik._tests.pools.test_pool_rpc import (
    LOCK_HEIGHT,
    create_new_plotnft,
    manage_temporary_pool_plot,
    verify_pool_state,
)
from chik._tests.util.misc import Marks, boolean_datacases, datacases
from chik.cmds.cmd_classes import ChikCliContext
from chik.cmds.cmd_helpers import NeedsWalletRPC, WalletClientInfo
from chik.cmds.param_types import CliAddress
from chik.cmds.plotnft import (
    ChangePayoutInstructionsPlotNFTCMD,
    ClaimPlotNFTCMD,
    CreatePlotNFTCMD,
    GetLoginLinkCMD,
    InspectPlotNFTCMD,
    JoinPlotNFTCMD,
    LeavePlotNFTCMD,
    ShowPlotNFTCMD,
)
from chik.pools.pool_config import PoolWalletConfig, load_pool_config, update_pool_config
from chik.pools.pool_wallet_info import PoolSingletonState, PoolWalletInfo
from chik.rpc.wallet_rpc_client import WalletRpcClient
from chik.simulator.setup_services import setup_farmer
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.bech32m import encode_puzzle_hash
from chik.util.config import lock_and_load_config, save_config
from chik.util.errors import CliRpcConnectionError
from chik.util.ints import uint64
from chik.wallet.util.address_type import AddressType
from chik.wallet.util.wallet_types import WalletType
from chik.wallet.wallet_state_manager import WalletStateManager

# limit to plain consensus mode for all tests
pytestmark = [pytest.mark.limit_consensus_modes(reason="irrelevant")]


@dataclass
class StateUrlCase:
    id: str
    state: str
    pool_url: Optional[str]
    expected_error: Optional[str] = None
    marks: Marks = ()


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [1],
        }
    ],
    indirect=True,
)
@boolean_datacases(name="self_pool", true="local", false="pool")
@boolean_datacases(name="prompt", true="prompt", false="dont_prompt")
@pytest.mark.anyio
async def test_plotnft_cli_create(
    wallet_environments: WalletTestFramework,
    self_pool: bool,
    prompt: bool,
    mocker: MockerFixture,
) -> None:
    wallet_state_manager: WalletStateManager = wallet_environments.environments[0].wallet_state_manager
    wallet_rpc: WalletRpcClient = wallet_environments.environments[0].rpc_client
    client_info: WalletClientInfo = WalletClientInfo(
        wallet_rpc,
        wallet_state_manager.root_pubkey.get_fingerprint(),
        wallet_state_manager.config,
    )

    wallet_state_manager.config["reuse_public_key_for_change"][str(client_info.fingerprint)] = (
        wallet_environments.tx_config.reuse_puzhash
    )

    state = "local" if self_pool else "pool"
    pool_url = None if self_pool else "http://pool.example.com"

    if not self_pool:
        pool_response_dict = {
            "name": "Pool Name",
            "description": "Pool Description",
            "logo_url": "https://subdomain.pool-domain.tld/path/to/logo.svg",
            "target_puzzle_hash": "344587cf06a39db471d2cc027504e8688a0a67cce961253500c956c73603fd58",
            "fee": "0.01",
            "protocol_version": 1,
            "relative_lock_height": 5,
            "minimum_difficulty": 1,
            "authentication_token_timeout": 5,
        }

        mock_get = mocker.patch("aiohttp.ClientSession.get")
        mock_get.return_value.__aenter__.return_value.text.return_value = json.dumps(pool_response_dict)

    if prompt:
        mocker.patch("sys.stdin", StringIO("yes\n"))

    await CreatePlotNFTCMD(
        rpc_info=NeedsWalletRPC(
            client_info=client_info,
        ),
        state=state,
        dont_prompt=not prompt,
        pool_url=pool_url,
    ).run()

    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                pre_block_balance_updates={
                    1: {
                        "confirmed_wallet_balance": 0,
                        "unconfirmed_wallet_balance": -1,
                        "<=#spendable_balance": 1,
                        "<=#max_send_amount": 1,
                        ">=#pending_change": 1,  # any amount increase
                        "pending_coin_removal_count": 1,
                    },
                },
                post_block_balance_updates={
                    1: {
                        "confirmed_wallet_balance": -1,
                        "unconfirmed_wallet_balance": 0,
                        ">=#spendable_balance": 1,
                        ">=#max_send_amount": 1,
                        "<=#pending_change": 1,  # any amount decrease
                        "<=#pending_coin_removal_count": 1,
                    },
                    2: {"init": True, "unspent_coin_count": 1},
                },
            )
        ]
    )

    summaries_response = await wallet_rpc.get_wallets(WalletType.POOLING_WALLET)
    assert len(summaries_response) == 1
    wallet_id: int = summaries_response[0]["id"]

    await verify_pool_state(wallet_rpc, wallet_id, PoolSingletonState.SELF_POOLING)


@datacases(
    StateUrlCase(
        id="local state with pool url",
        state="local",
        pool_url="https://pool.example.com",
        expected_error="is not allowed with 'local' state",
    ),
    StateUrlCase(
        id="pool state no pool url",
        state="pool",
        pool_url=None,
        expected_error="is required with 'pool' state",
    ),
)
@pytest.mark.anyio
async def test_plotnft_cli_create_errors(
    case: StateUrlCase,
    consensus_mode: ConsensusMode,
) -> None:
    with pytest.raises(CliRpcConnectionError, match=case.expected_error):
        await CreatePlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=None,
                wallet_rpc_port=None,
                fingerprint=None,
            ),
            state=case.state,
            dont_prompt=True,
            pool_url=case.pool_url,
        ).run()


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [1],
        }
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_plotnft_cli_show(
    wallet_environments: WalletTestFramework,
    capsys: pytest.CaptureFixture[str],
) -> None:
    wallet_state_manager: WalletStateManager = wallet_environments.environments[0].wallet_state_manager
    wallet_rpc: WalletRpcClient = wallet_environments.environments[0].rpc_client
    client_info: WalletClientInfo = WalletClientInfo(
        wallet_rpc,
        wallet_state_manager.root_pubkey.get_fingerprint(),
        wallet_state_manager.config,
    )
    root_path = wallet_environments.environments[0].node.root_path
    wallet_state_manager.config["reuse_public_key_for_change"][str(client_info.fingerprint)] = (
        wallet_environments.tx_config.reuse_puzhash
    )

    await ShowPlotNFTCMD(
        # we need this for the farmer rpc client which is used in the comment
        context=ChikCliContext(root_path=root_path),
        rpc_info=NeedsWalletRPC(
            client_info=client_info,
        ),
        id=None,
    ).run()
    out, _err = capsys.readouterr()
    assert "Wallet height: 3\nSync status: Synced\n" == out

    with pytest.raises(CliRpcConnectionError, match="is not a pool wallet"):
        await ShowPlotNFTCMD(
            context=ChikCliContext(root_path=root_path),
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=15,
        ).run()

    wallet_id = await create_new_plotnft(wallet_environments)

    # need to capture the output and verify
    await ShowPlotNFTCMD(
        context=ChikCliContext(root_path=root_path),
        rpc_info=NeedsWalletRPC(
            client_info=client_info,
        ),
        id=wallet_id,
    ).run()
    out, _err = capsys.readouterr()
    assert "Current state: FARMING_TO_POOL" in out
    assert f"Wallet ID: {wallet_id}" in out

    wallet_id_2 = await create_new_plotnft(wallet_environments, self_pool=False, second_nft=True)

    # Passing in None when there are multiple pool wallets
    # Should show the state of all pool wallets
    await ShowPlotNFTCMD(
        context=ChikCliContext(root_path=root_path),
        rpc_info=NeedsWalletRPC(
            client_info=client_info,
        ),
        id=None,
    ).run()
    out, _err = capsys.readouterr()
    assert "Current state: FARMING_TO_POOL" in out
    assert f"Wallet ID: {wallet_id}" in out
    assert f"Wallet ID: {wallet_id_2}" in out


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [1],
        }
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_plotnft_cli_show_with_farmer(
    wallet_environments: WalletTestFramework,
    capsys: pytest.CaptureFixture[str],
    self_hostname: str,
    # with_wallet_id: bool,
) -> None:
    wallet_state_manager: WalletStateManager = wallet_environments.environments[0].wallet_state_manager
    wallet_rpc: WalletRpcClient = wallet_environments.environments[0].rpc_client
    client_info: WalletClientInfo = WalletClientInfo(
        wallet_rpc,
        wallet_state_manager.root_pubkey.get_fingerprint(),
        wallet_state_manager.config,
    )
    wallet_state_manager.config["reuse_public_key_for_change"][str(client_info.fingerprint)] = (
        wallet_environments.tx_config.reuse_puzhash
    )

    #  Need to run the farmer to make further tests
    root_path = wallet_environments.environments[0].node.root_path

    async with setup_farmer(
        b_tools=wallet_environments.full_node.bt,
        root_path=root_path,
        self_hostname=self_hostname,
        consensus_constants=wallet_environments.full_node.bt.constants,
    ) as farmer:
        assert farmer.rpc_server and farmer.rpc_server.webserver

        with lock_and_load_config(root_path, "config.yaml") as config:
            config["farmer"]["rpc_port"] = farmer.rpc_server.webserver.listen_port
            save_config(root_path, "config.yaml", config)

        await ShowPlotNFTCMD(
            context=ChikCliContext(root_path=root_path),
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=None,
        ).run()
        out, _err = capsys.readouterr()
        assert "Sync status: Synced" in out
        assert "Current state" not in out

        wallet_id = await create_new_plotnft(wallet_environments)
        pw_info, _ = await wallet_rpc.pw_status(wallet_id)

        await ShowPlotNFTCMD(
            context=ChikCliContext(root_path=root_path),
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=wallet_id,
        ).run()
        out, _err = capsys.readouterr()
        assert "Current state: FARMING_TO_POOL" in out
        assert f"Wallet ID: {wallet_id}" in out
        assert f"Launcher ID: {pw_info.launcher_id.hex()}" in out


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [10],
        }
    ],
    indirect=True,
)
@boolean_datacases(name="prompt", true="prompt", false="dont_prompt")
@pytest.mark.anyio
async def test_plotnft_cli_leave(
    wallet_environments: WalletTestFramework,
    prompt: bool,
    mocker: MockerFixture,
) -> None:
    wallet_state_manager: WalletStateManager = wallet_environments.environments[0].wallet_state_manager
    wallet_rpc: WalletRpcClient = wallet_environments.environments[0].rpc_client
    client_info: WalletClientInfo = WalletClientInfo(
        wallet_rpc,
        wallet_state_manager.root_pubkey.get_fingerprint(),
        wallet_state_manager.config,
    )
    wallet_state_manager.config["reuse_public_key_for_change"][str(client_info.fingerprint)] = (
        wallet_environments.tx_config.reuse_puzhash
    )

    if prompt:
        mocker.patch("sys.stdin", StringIO("yes\n"))

    with pytest.raises(CliRpcConnectionError, match="No pool wallet found"):
        await LeavePlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=None,
            dont_prompt=not prompt,
        ).run()

    with pytest.raises(CliRpcConnectionError, match="is not a pool wallet"):
        await LeavePlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=15,
            dont_prompt=not prompt,
        ).run()

    wallet_id = await create_new_plotnft(wallet_environments)

    await LeavePlotNFTCMD(
        rpc_info=NeedsWalletRPC(
            client_info=client_info,
        ),
        id=wallet_id,
        dont_prompt=not prompt,
    ).run()

    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                pre_block_balance_updates={
                    1: {
                        "<=#spendable_balance": 1,
                        "<=#max_send_amount": 1,
                        "pending_coin_removal_count": 0,
                    },
                    2: {"pending_coin_removal_count": 1},
                },
                post_block_balance_updates={
                    1: {
                        "<=#pending_coin_removal_count": 1,
                    },
                    2: {"pending_coin_removal_count": -1},
                },
            )
        ]
    )

    await verify_pool_state(wallet_rpc, wallet_id, PoolSingletonState.LEAVING_POOL)

    await wallet_environments.full_node.farm_blocks_to_puzzlehash(
        count=LOCK_HEIGHT + 2, guarantee_transaction_blocks=True
    )

    await verify_pool_state(wallet_rpc, wallet_id, PoolSingletonState.SELF_POOLING)


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [10],
        }
    ],
    indirect=True,
)
@boolean_datacases(name="prompt", true="prompt", false="dont_prompt")
@pytest.mark.anyio
async def test_plotnft_cli_join(
    wallet_environments: WalletTestFramework,
    prompt: bool,
    mocker: MockerFixture,
) -> None:
    wallet_state_manager: WalletStateManager = wallet_environments.environments[0].wallet_state_manager
    wallet_rpc: WalletRpcClient = wallet_environments.environments[0].rpc_client
    client_info: WalletClientInfo = WalletClientInfo(
        wallet_rpc,
        wallet_state_manager.root_pubkey.get_fingerprint(),
        wallet_state_manager.config,
    )
    wallet_state_manager.config["reuse_public_key_for_change"][str(client_info.fingerprint)] = (
        wallet_environments.tx_config.reuse_puzhash
    )

    # Test error cases
    # No pool wallet found
    with pytest.raises(CliRpcConnectionError, match="No pool wallet found"):
        await JoinPlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            pool_url="http://127.0.0.1",
            id=None,
            dont_prompt=not prompt,
        ).run()

    # Wallet id not a pool wallet
    with pytest.raises(CliRpcConnectionError, match="is not a pool wallet"):
        await JoinPlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            pool_url="http://127.0.0.1",
            id=1,
            dont_prompt=not prompt,
        ).run()

    # Create a farming plotnft to url http://pool.example.com
    wallet_id = await create_new_plotnft(wallet_environments)

    # HTTPS check on mainnet
    with pytest.raises(CliRpcConnectionError, match="must be HTTPS on mainnet"):
        config_override = wallet_state_manager.config.copy()
        config_override["selected_network"] = "mainnet"
        mainnet_override = WalletClientInfo(client_info.client, client_info.fingerprint, config_override)
        await JoinPlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=mainnet_override,
            ),
            pool_url="http://127.0.0.1",
            id=wallet_id,
            dont_prompt=not prompt,
        ).run()

    # Some more error cases
    with pytest.raises(CliRpcConnectionError, match="Error connecting to pool"):
        await JoinPlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=wallet_id,
            pool_url="http://127.0.0.1",
            dont_prompt=not prompt,
        ).run()

    with pytest.raises(CliRpcConnectionError, match="Error connecting to pool"):
        await JoinPlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=wallet_id,
            pool_url="",
            dont_prompt=not prompt,
        ).run()

    pool_response_dict = {
        "name": "Pool Name",
        "description": "Pool Description",
        "logo_url": "https://subdomain.pool-domain.tld/path/to/logo.svg",
        "target_puzzle_hash": "344587cf06a39db471d2cc027504e8688a0a67cce961253500c956c73603fd58",
        "fee": "0.01",
        "protocol_version": 1,
        "relative_lock_height": 50000,
        "minimum_difficulty": 1,
        "authentication_token_timeout": 5,
    }

    mock_get = mocker.patch("aiohttp.ClientSession.get")
    mock_get.return_value.__aenter__.return_value.text.return_value = json.dumps(pool_response_dict)

    with pytest.raises(CliRpcConnectionError, match="Relative lock height too high for this pool"):
        await JoinPlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=wallet_id,
            pool_url="",
            dont_prompt=not prompt,
        ).run()

    pool_response_dict["relative_lock_height"] = LOCK_HEIGHT
    pool_response_dict["protocol_version"] = 2
    mock_get.return_value.__aenter__.return_value.text.return_value = json.dumps(pool_response_dict)

    with pytest.raises(CliRpcConnectionError, match="Incorrect version"):
        await JoinPlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=wallet_id,
            pool_url="",
            dont_prompt=not prompt,
        ).run()

    pool_response_dict["relative_lock_height"] = LOCK_HEIGHT
    pool_response_dict["protocol_version"] = 1
    mock_get.return_value.__aenter__.return_value.text.return_value = json.dumps(pool_response_dict)

    if prompt:
        mocker.patch("sys.stdin", StringIO("yes\n"))

    # Join the new pool - this will leave the prior pool and join the new one
    # Here you can use None as the wallet_id and the code will pick the only pool wallet automatically
    await JoinPlotNFTCMD(
        rpc_info=NeedsWalletRPC(
            client_info=client_info,
        ),
        id=None,
        pool_url="http://127.0.0.1",
        dont_prompt=not prompt,
    ).run()

    await wallet_environments.full_node.farm_blocks_to_puzzlehash(count=1, guarantee_transaction_blocks=True)
    await verify_pool_state(wallet_rpc, wallet_id, PoolSingletonState.LEAVING_POOL)
    await wallet_environments.full_node.farm_blocks_to_puzzlehash(
        count=LOCK_HEIGHT + 2, guarantee_transaction_blocks=True
    )
    await verify_pool_state(wallet_rpc, wallet_id, PoolSingletonState.FARMING_TO_POOL)
    await wallet_environments.full_node.wait_for_wallet_synced(
        wallet_node=wallet_environments.environments[0].node, timeout=20
    )

    # Create a second farming plotnft to url http://pool.example.com
    wallet_id = await create_new_plotnft(wallet_environments, self_pool=False, second_nft=True)

    # Join the new pool - this will leave the prior pool and join the new one
    # Will fail because we don't specify a wallet ID and there are multiple pool wallets
    with pytest.raises(CliRpcConnectionError, match="More than one pool wallet"):
        await JoinPlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=None,
            pool_url="http://127.0.0.1",
            dont_prompt=not prompt,
        ).run()

    if prompt:
        mocker.patch("sys.stdin", StringIO("yes\n"))

    # Join the new pool - this will leave the prior pool and join the new one and specific wallet_id
    await JoinPlotNFTCMD(
        rpc_info=NeedsWalletRPC(
            client_info=client_info,
        ),
        id=wallet_id,
        pool_url="http://127.0.0.1",
        dont_prompt=not prompt,
    ).run()

    await wallet_environments.full_node.farm_blocks_to_puzzlehash(count=1, guarantee_transaction_blocks=True)
    await verify_pool_state(wallet_rpc, wallet_id, PoolSingletonState.LEAVING_POOL)
    await wallet_environments.full_node.farm_blocks_to_puzzlehash(
        count=LOCK_HEIGHT + 2, guarantee_transaction_blocks=True
    )
    await verify_pool_state(wallet_rpc, wallet_id, PoolSingletonState.FARMING_TO_POOL)

    # Join the same pool test - code not ready yet for test
    # Needs PR #18822
    # with pytest.raises(CliRpcConnectionError, match="already joined"):
    #     await JoinPlotNFTCMD(
    #         rpc_info=NeedsWalletRPC(
    #             client_info=client_info,
    #         ),
    #         id=wallet_id,
    #         pool_url="http://127.0.0.1",
    #         dont_prompt=not prompt,
    #     ).run()


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [10],
        }
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_plotnft_cli_claim(
    wallet_environments: WalletTestFramework,
) -> None:
    wallet_state_manager: WalletStateManager = wallet_environments.environments[0].wallet_state_manager
    wallet_rpc: WalletRpcClient = wallet_environments.environments[0].rpc_client
    client_info: WalletClientInfo = WalletClientInfo(
        wallet_rpc,
        wallet_state_manager.root_pubkey.get_fingerprint(),
        wallet_state_manager.config,
    )
    wallet_state_manager.config["reuse_public_key_for_change"][str(client_info.fingerprint)] = (
        wallet_environments.tx_config.reuse_puzhash
    )

    # Test error cases
    # No pool wallet found
    with pytest.raises(CliRpcConnectionError, match="No pool wallet found"):
        await ClaimPlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=None,
        ).run()

    # Wallet id not a pool wallet
    with pytest.raises(CliRpcConnectionError, match="is not a pool wallet"):
        await ClaimPlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=1,
        ).run()

    # Create a self-pooling plotnft
    wallet_id = await create_new_plotnft(wallet_environments, self_pool=True)

    status: PoolWalletInfo = (await wallet_rpc.pw_status(wallet_id))[0]
    our_ph = await wallet_state_manager.main_wallet.get_new_puzzlehash()
    bt = wallet_environments.full_node.bt

    async with manage_temporary_pool_plot(bt, status.p2_singleton_puzzle_hash) as pool_plot:
        all_blocks = await wallet_environments.full_node.get_all_full_blocks()
        blocks = bt.get_consecutive_blocks(
            3,
            block_list_input=all_blocks,
            force_plot_id=pool_plot.plot_id,
            farmer_reward_puzzle_hash=our_ph,
            guarantee_transaction_block=True,
        )

    for block in blocks[-3:]:
        await wallet_environments.full_node.full_node.add_block(block)

    await wallet_environments.full_node.wait_for_wallet_synced(
        wallet_node=wallet_environments.environments[0].node, timeout=20
    )
    await ClaimPlotNFTCMD(
        rpc_info=NeedsWalletRPC(
            client_info=client_info,
        ),
        id=None,
    ).run()

    await wallet_environments.process_pending_states(
        [
            WalletStateTransition(
                pre_block_balance_updates={
                    1: {
                        "confirmed_wallet_balance": 500_000_000_000,
                        "unconfirmed_wallet_balance": 500_000_000_000,
                        "spendable_balance": 500_000_000_000,
                        "max_send_amount": 500_000_000_000,
                        "pending_change": 0,
                        "unspent_coin_count": 2,
                        "pending_coin_removal_count": 0,
                    },
                    2: {
                        "confirmed_wallet_balance": 2 * 1_750_000_000_000,
                        "unconfirmed_wallet_balance": 2 * 1_750_000_000_000,
                        "spendable_balance": 2 * 1_750_000_000_000,
                        "max_send_amount": 0,
                        "pending_change": 0,
                        "unspent_coin_count": 2,
                        "pending_coin_removal_count": 3,
                    },
                },
                post_block_balance_updates={
                    1: {
                        "confirmed_wallet_balance": +3_750_000_000_000,  # two pool rewards and 1 farm reward
                        "unconfirmed_wallet_balance": +3_750_000_000_000,
                        "spendable_balance": +3_750_000_000_000,
                        "max_send_amount": +3_750_000_000_000,
                        "pending_change": 0,
                        "unspent_coin_count": +3,
                        "pending_coin_removal_count": 0,
                    },
                    2: {
                        "confirmed_wallet_balance": -1_750_000_000_000,
                        "unconfirmed_wallet_balance": -1_750_000_000_000,
                        "spendable_balance": -1_750_000_000_000,
                        "max_send_amount": 0,
                        "pending_change": 0,
                        "unspent_coin_count": -1,
                        "pending_coin_removal_count": -3,
                    },
                },
            )
        ]
    )


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [10],
        }
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_plotnft_cli_inspect(
    wallet_environments: WalletTestFramework,
    capsys: pytest.CaptureFixture[str],
) -> None:
    wallet_state_manager: WalletStateManager = wallet_environments.environments[0].wallet_state_manager
    wallet_rpc: WalletRpcClient = wallet_environments.environments[0].rpc_client
    client_info: WalletClientInfo = WalletClientInfo(
        wallet_rpc,
        wallet_state_manager.root_pubkey.get_fingerprint(),
        wallet_state_manager.config,
    )
    wallet_state_manager.config["reuse_public_key_for_change"][str(client_info.fingerprint)] = (
        wallet_environments.tx_config.reuse_puzhash
    )

    with pytest.raises(CliRpcConnectionError, match="No pool wallet found"):
        await InspectPlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=None,
        ).run()

    with pytest.raises(CliRpcConnectionError, match="is not a pool wallet"):
        await InspectPlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=15,
        ).run()

    wallet_id = await create_new_plotnft(wallet_environments)

    # need to capture the output and verify
    await InspectPlotNFTCMD(
        rpc_info=NeedsWalletRPC(
            client_info=client_info,
        ),
        id=wallet_id,
    ).run()
    out, _err = capsys.readouterr()
    json_output = json.loads(out)

    assert (
        json_output["pool_wallet_info"]["current"]["owner_pubkey"]
        == "0xb286bbf7a10fa058d2a2a758921377ef00bb7f8143e1bd40dd195ae918dbef42cfc481140f01b9eae13b430a0c8fe304"
    )
    assert json_output["pool_wallet_info"]["current"]["state"] == PoolSingletonState.FARMING_TO_POOL.value

    wallet_id = await create_new_plotnft(wallet_environments, self_pool=True, second_nft=True)

    with pytest.raises(CliRpcConnectionError, match="More than one pool wallet"):
        await InspectPlotNFTCMD(
            rpc_info=NeedsWalletRPC(
                client_info=client_info,
            ),
            id=None,
        ).run()

    await InspectPlotNFTCMD(
        rpc_info=NeedsWalletRPC(
            client_info=client_info,
        ),
        id=wallet_id,
    ).run()
    out, _err = capsys.readouterr()
    json_output = json.loads(out)

    assert (
        json_output["pool_wallet_info"]["current"]["owner_pubkey"]
        == "0x893474c97d04a0283483ba1af9e070768dff9e9a83d9ae2cf00a34be96ca29aec387dfb7474f2548d777000e5463f602"
    )

    assert json_output["pool_wallet_info"]["current"]["state"] == PoolSingletonState.SELF_POOLING.value


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [10],
        }
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_plotnft_cli_change_payout(
    wallet_environments: WalletTestFramework,
    mocker: MockerFixture,
    capsys: pytest.CaptureFixture[str],
) -> None:
    wallet_state_manager: WalletStateManager = wallet_environments.environments[0].wallet_state_manager
    wallet_rpc: WalletRpcClient = wallet_environments.environments[0].rpc_client
    client_info: WalletClientInfo = WalletClientInfo(
        wallet_rpc,
        wallet_state_manager.root_pubkey.get_fingerprint(),
        wallet_state_manager.config,
    )
    wallet_state_manager.config["reuse_public_key_for_change"][str(client_info.fingerprint)] = (
        wallet_environments.tx_config.reuse_puzhash
    )

    zero_ph = bytes32.from_hexstr("0x0000000000000000000000000000000000000000000000000000000000000000")
    zero_address = encode_puzzle_hash(zero_ph, "xck")

    burn_ph = bytes32.from_hexstr("0x000000000000000000000000000000000000000000000000000000000000dead")
    burn_address = encode_puzzle_hash(burn_ph, "xck")
    root_path = wallet_environments.environments[0].node.root_path

    wallet_id = await create_new_plotnft(wallet_environments)
    pw_info, _ = await wallet_rpc.pw_status(wallet_id)

    # This tests what happens when using None for root_path
    mocker.patch("chik.cmds.plotnft_funcs.DEFAULT_ROOT_PATH", root_path)
    await ChangePayoutInstructionsPlotNFTCMD(
        context=ChikCliContext(root_path=wallet_environments.environments[0].node.root_path),
        launcher_id=bytes32(32 * b"0"),
        address=CliAddress(burn_ph, burn_address, AddressType.XCK),
    ).run()
    out, _err = capsys.readouterr()
    assert f"{bytes32(32 * b'0').hex()} Not found." in out

    new_config: PoolWalletConfig = PoolWalletConfig(
        launcher_id=pw_info.launcher_id,
        pool_url="http://pool.example.com",
        payout_instructions=zero_address,
        target_puzzle_hash=bytes32(32 * b"0"),
        p2_singleton_puzzle_hash=pw_info.p2_singleton_puzzle_hash,
        owner_public_key=G1Element(),
    )

    await update_pool_config(root_path=root_path, pool_config_list=[new_config])
    config: list[PoolWalletConfig] = load_pool_config(root_path)
    wanted_config = next((x for x in config if x.launcher_id == pw_info.launcher_id), None)
    assert wanted_config is not None
    assert wanted_config.payout_instructions == zero_address

    await ChangePayoutInstructionsPlotNFTCMD(
        context=ChikCliContext(root_path=root_path),
        launcher_id=pw_info.launcher_id,
        address=CliAddress(burn_ph, burn_address, AddressType.XCK),
    ).run()
    out, _err = capsys.readouterr()
    assert f"Payout Instructions for launcher id: {pw_info.launcher_id.hex()} successfully updated" in out

    config = load_pool_config(root_path)
    wanted_config = next((x for x in config if x.launcher_id == pw_info.launcher_id), None)
    assert wanted_config is not None
    assert wanted_config.payout_instructions == burn_ph.hex()


@pytest.mark.parametrize(
    "wallet_environments",
    [
        {
            "num_environments": 1,
            "blocks_needed": [10],
        }
    ],
    indirect=True,
)
@pytest.mark.anyio
async def test_plotnft_cli_get_login_link(
    capsys: pytest.CaptureFixture[str],
    wallet_environments: WalletTestFramework,
    self_hostname: str,
) -> None:
    wallet_state_manager: WalletStateManager = wallet_environments.environments[0].wallet_state_manager
    wallet_rpc: WalletRpcClient = wallet_environments.environments[0].rpc_client
    _client_info: WalletClientInfo = WalletClientInfo(
        wallet_rpc,
        wallet_state_manager.root_pubkey.get_fingerprint(),
        wallet_state_manager.config,
    )
    bt = wallet_environments.full_node.bt

    async with setup_farmer(
        b_tools=bt,
        root_path=wallet_environments.environments[0].node.root_path,
        self_hostname=self_hostname,
        consensus_constants=bt.constants,
    ) as farmer:
        root_path = wallet_environments.environments[0].node.root_path

        assert farmer.rpc_server and farmer.rpc_server.webserver
        with lock_and_load_config(root_path, "config.yaml") as config:
            config["farmer"]["rpc_port"] = farmer.rpc_server.webserver.listen_port
            save_config(root_path, "config.yaml", config)
        with pytest.raises(CliRpcConnectionError, match="Was not able to get login link"):
            await GetLoginLinkCMD(
                context=ChikCliContext(root_path=root_path),
                launcher_id=bytes32(32 * b"0"),
            ).run()


@pytest.mark.anyio
async def test_plotnft_cli_misc(mocker: MockerFixture, consensus_mode: ConsensusMode) -> None:
    from chik.cmds.plotnft_funcs import create

    test_rpc_client = TestWalletRpcClient()

    with pytest.raises(CliRpcConnectionError, match="Pool URLs must be HTTPS on mainnet"):
        await create(
            wallet_info=WalletClientInfo(
                client=cast(WalletRpcClient, test_rpc_client),
                fingerprint=0,
                config={"selected_network": "mainnet"},
            ),
            pool_url="http://pool.example.com",
            state="FARMING_TO_POOL",
            fee=uint64(0),
            prompt=False,
        )

    with pytest.raises(ValueError, match="Plot NFT must be created in SELF_POOLING or FARMING_TO_POOL state"):
        await create(
            wallet_info=WalletClientInfo(client=cast(WalletRpcClient, test_rpc_client), fingerprint=0, config=dict()),
            pool_url=None,
            state="Invalid State",
            fee=uint64(0),
            prompt=False,
        )

    # Test fall-through raise in create
    mocker.patch.object(
        test_rpc_client, "create_new_pool_wallet", create=True, side_effect=ValueError("Injected error")
    )
    with pytest.raises(CliRpcConnectionError, match="Error creating plot NFT: Injected error"):
        await create(
            wallet_info=WalletClientInfo(client=cast(WalletRpcClient, test_rpc_client), fingerprint=0, config=dict()),
            pool_url=None,
            state="SELF_POOLING",
            fee=uint64(0),
            prompt=False,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.serialized_program import SerializedProgram


@dataclass(frozen=True)
class UncurriedPuzzle:
    mod: Program
    args: Program


def uncurry_puzzle(puzzle: Union[Program, SerializedProgram]) -> UncurriedPuzzle:
    return UncurriedPuzzle(*puzzle.uncurry())



from __future__ import annotations

import random
import time

import pytest
from chik_rs import G2Element

from chik._tests.util.db_connection import DBConnection
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint32, uint64
from chik.wallet.conditions import ConditionValidTimes
from chik.wallet.trade_record import TradeRecord, TradeRecordOld
from chik.wallet.trading.offer import Offer
from chik.wallet.trading.trade_status import TradeStatus
from chik.wallet.trading.trade_store import TradeStore, migrate_coin_of_interest
from chik.wallet.util.wallet_types import WalletType
from chik.wallet.wallet_coin_record import WalletCoinRecord
from chik.wallet.wallet_coin_store import WalletCoinStore
from chik.wallet.wallet_spend_bundle import WalletSpendBundle

module_seeded_random = random.Random()
module_seeded_random.seed(a=0, version=2)

coin_1 = Coin(bytes32.random(module_seeded_random), bytes32.random(module_seeded_random), uint64(12311))
coin_2 = Coin(coin_1.parent_coin_info, bytes32.random(module_seeded_random), uint64(12312))
coin_3 = Coin(coin_1.parent_coin_info, bytes32.random(module_seeded_random), uint64(12313))
record_1 = WalletCoinRecord(coin_1, uint32(4), uint32(0), False, True, WalletType.STANDARD_WALLET, 0)
record_2 = WalletCoinRecord(coin_2, uint32(5), uint32(0), False, True, WalletType.STANDARD_WALLET, 0)
record_3 = WalletCoinRecord(coin_3, uint32(6), uint32(0), False, True, WalletType.STANDARD_WALLET, 0)


@pytest.mark.anyio
async def test_get_coins_of_interest_with_trade_statuses(seeded_random: random.Random) -> None:
    async with DBConnection(1) as db_wrapper:
        coin_store = await WalletCoinStore.create(db_wrapper)
        trade_store = await TradeStore.create(db_wrapper)
        await coin_store.add_coin_record(record_1)
        await coin_store.add_coin_record(record_2)
        await coin_store.add_coin_record(record_3)

        tr1_name: bytes32 = bytes32.random(seeded_random)
        tr1 = TradeRecord(
            confirmed_at_index=uint32(0),
            accepted_at_time=None,
            created_at_time=uint64(time.time()),
            is_my_offer=True,
            sent=uint32(0),
            offer=bytes([1, 2, 3]),
            taken_offer=None,
            coins_of_interest=[coin_2],
            trade_id=tr1_name,
            status=uint32(TradeStatus.PENDING_ACCEPT.value),
            sent_to=[],
            valid_times=ConditionValidTimes(),
        )
        await trade_store.add_trade_record(tr1, offer_name=bytes32.random(seeded_random))

        tr2_name: bytes32 = bytes32.random(seeded_random)
        tr2 = TradeRecord(
            confirmed_at_index=uint32(0),
            accepted_at_time=None,
            created_at_time=uint64(time.time()),
            is_my_offer=True,
            sent=uint32(0),
            offer=bytes([1, 2, 3]),
            taken_offer=None,
            coins_of_interest=[coin_1, coin_3],
            trade_id=tr2_name,
            status=uint32(TradeStatus.PENDING_CONFIRM.value),
            sent_to=[],
            valid_times=ConditionValidTimes(),
        )
        await trade_store.add_trade_record(tr2, offer_name=bytes32.random(seeded_random))

        assert await trade_store.get_coin_ids_of_interest_with_trade_statuses([TradeStatus.PENDING_CONFIRM]) == {
            coin_1.name(),
            coin_3.name(),
        }
        assert await trade_store.get_coin_ids_of_interest_with_trade_statuses([TradeStatus.PENDING_ACCEPT]) == {
            coin_2.name()
        }

        # test replace trade record
        tr2_1 = TradeRecord(
            confirmed_at_index=uint32(0),
            accepted_at_time=None,
            created_at_time=uint64(time.time()),
            is_my_offer=True,
            sent=uint32(0),
            offer=bytes([1, 2, 3]),
            taken_offer=None,
            coins_of_interest=[coin_2],
            trade_id=tr2_name,
            status=uint32(TradeStatus.PENDING_CONFIRM.value),
            sent_to=[],
            valid_times=ConditionValidTimes(),
        )
        await trade_store.add_trade_record(tr2_1, offer_name=bytes32.random(seeded_random))

        assert await trade_store.get_coin_ids_of_interest_with_trade_statuses([TradeStatus.PENDING_CONFIRM]) == {
            coin_2.name()
        }

        # test migration
        async with trade_store.db_wrapper.writer_maybe_transaction() as conn:
            await conn.execute("DELETE FROM coin_of_interest_to_trade_record")

        assert await trade_store.get_coin_ids_of_interest_with_trade_statuses([TradeStatus.PENDING_ACCEPT]) == set()

        async with trade_store.db_wrapper.writer_maybe_transaction() as conn:
            await migrate_coin_of_interest(trade_store.log, conn)

        assert await trade_store.get_coin_ids_of_interest_with_trade_statuses([TradeStatus.PENDING_ACCEPT]) == {
            coin_2.name()
        }


@pytest.mark.anyio
async def test_valid_times_migration() -> None:
    async with DBConnection(1) as db_wrapper:
        async with db_wrapper.writer_maybe_transaction() as conn:
            await conn.execute(
                "CREATE TABLE IF NOT EXISTS trade_records("
                " trade_record blob,"
                " trade_id text PRIMARY KEY,"
                " status int,"
                " confirmed_at_index int,"
                " created_at_time bigint,"
                " sent int,"
                " is_my_offer tinyint)"
            )

        fake_offer = Offer({}, WalletSpendBundle([], G2Element()), {})
        fake_coin = Coin(bytes32.zeros, bytes32.zeros, uint64(0))
        old_record = TradeRecordOld(
            confirmed_at_index=uint32(0),
            accepted_at_time=None,
            created_at_time=uint64(1000000),
            is_my_offer=True,
            sent=uint32(0),
            offer=bytes(fake_offer),
            taken_offer=None,
            coins_of_interest=[fake_coin],
            trade_id=bytes32.zeros,
            status=uint32(TradeStatus.PENDING_ACCEPT.value),
            sent_to=[],
        )

        async with db_wrapper.writer_maybe_transaction() as conn:
            cursor = await conn.execute(
                "INSERT INTO trade_records "
                "(trade_record, trade_id, status, confirmed_at_index, created_at_time, sent, is_my_offer) "
                "VALUES(?, ?, ?, ?, ?, ?, ?)",
                (
                    bytes(old_record),
                    old_record.trade_id.hex(),
                    old_record.status,
                    old_record.confirmed_at_index,
                    old_record.created_at_time,
                    old_record.sent,
                    old_record.is_my_offer,
                ),
            )
            await cursor.close()

        trade_store = await TradeStore.create(db_wrapper)
        rec = await trade_store.get_trade_record(old_record.trade_id)
        assert rec is not None
        assert rec.valid_times == ConditionValidTimes()


@pytest.mark.anyio
async def test_large_trade_record_query() -> None:
    async with DBConnection(1) as db_wrapper:
        store = await TradeStore.create(db_wrapper)
        trade_records_to_insert = []
        for _ in range(db_wrapper.host_parameter_limit + 1):
            offer_name = bytes32.random()
            trade_record_old = TradeRecordOld(
                confirmed_at_index=uint32(0),
                accepted_at_time=None,
                created_at_time=uint64(1000000),
                is_my_offer=True,
                sent=uint32(0),
                offer=b"",
                taken_offer=None,
                coins_of_interest=[],
                trade_id=offer_name,
                status=uint32(TradeStatus.PENDING_ACCEPT.value),
                sent_to=[],
            )
            trade_records_to_insert.append(
                (
                    bytes(trade_record_old),
                    trade_record_old.trade_id.hex(),
                    trade_record_old.status,
                    trade_record_old.confirmed_at_index,
                    trade_record_old.created_at_time,
                    trade_record_old.sent,
                    trade_record_old.trade_id,
                    trade_record_old.is_my_offer,
                )
            )
        async with db_wrapper.writer_maybe_transaction() as conn:
            await conn.executemany("INSERT INTO trade_records VALUES(?, ?, ?, ?, ?, ?, ?, ?)", trade_records_to_insert)
            # Insert a specific trade_record_times item for the last trade_records item
            await conn.execute(
                "INSERT INTO trade_record_times VALUES(?, ?)",
                (offer_name, bytes(ConditionValidTimes(min_height=uint32(42)))),
            )
        all_trades = await store.get_all_trades()
        assert len(all_trades) == db_wrapper.host_parameter_limit + 1
        # Check that all trade_record items have correct valid_times
        empty = ConditionValidTimes()
        assert all(trade.valid_times == empty for trade in all_trades[:-1])
        assert all_trades[-1].valid_times.min_height == uint32(42)



from __future__ import annotations

import pytest

from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.hash import std_hash
from chik.util.ints import uint64


def coin_serialize(amount: uint64, klvm_serialize: bytes, full_serialize: bytes):
    c = Coin(bytes32(b"a" * 32), bytes32(b"b" * 32), amount)
    expected_hash = (b"a" * 32) + (b"b" * 32) + klvm_serialize

    expected_serialization = (b"a" * 32) + (b"b" * 32) + full_serialize

    assert c.name() == std_hash(expected_hash)
    assert c.to_bytes() == expected_serialization
    assert bytes(c) == expected_serialization

    # make sure the serialization round-trips
    c2 = Coin.from_bytes(expected_serialization)
    assert c2 == c


def test_serialization():
    coin_serialize(uint64(0xFFFF), bytes([0, 0xFF, 0xFF]), bytes([0, 0, 0, 0, 0, 0, 0xFF, 0xFF]))
    coin_serialize(uint64(1337000000), bytes([0x4F, 0xB1, 0x00, 0x40]), bytes([0, 0, 0, 0, 0x4F, 0xB1, 0x00, 0x40]))

    # if the amount is 0, the amount is omitted in the "short" format,
    # that's hashed
    coin_serialize(uint64(0), b"", bytes([0, 0, 0, 0, 0, 0, 0, 0]))

    # when amount is > INT64_MAX, the "short" serialization format is 1 byte
    # longer, since it needs a leading zero to make it positive
    coin_serialize(
        uint64(0xFFFFFFFFFFFFFFFF),
        bytes([0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF]),
        bytes([0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF]),
    )


@pytest.mark.parametrize(
    "amount, klvm",
    [
        (0, []),
        (1, [1]),
        (0xFF, [0, 0xFF]),
        (0xFFFF, [0, 0xFF, 0xFF]),
        (0xFFFFFF, [0, 0xFF, 0xFF, 0xFF]),
        (0xFFFFFFFF, [0, 0xFF, 0xFF, 0xFF, 0xFF]),
        (0xFFFFFFFFFF, [0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF]),
        (0xFFFFFFFFFFFF, [0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF]),
        (0xFFFFFFFFFFFFFF, [0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF]),
        (0xFFFFFFFFFFFFFFFF, [0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF]),
        (0x7F, [0x7F]),
        (0x7FFF, [0x7F, 0xFF]),
        (0x7FFFFF, [0x7F, 0xFF, 0xFF]),
        (0x7FFFFFFF, [0x7F, 0xFF, 0xFF, 0xFF]),
        (0x7FFFFFFFFF, [0x7F, 0xFF, 0xFF, 0xFF, 0xFF]),
        (0x7FFFFFFFFFFF, [0x7F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF]),
        (0x7FFFFFFFFFFFFF, [0x7F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF]),
        (0x7FFFFFFFFFFFFFFF, [0x7F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF]),
    ],
)
def test_name(amount: int, klvm: list[int]) -> None:
    H1 = bytes32(b"a" * 32)
    H2 = bytes32(b"b" * 32)

    assert Coin(H1, H2, uint64(amount)).name() == std_hash(H1 + H2 + bytes(klvm))


def test_construction() -> None:
    H1 = b"a" * 32
    H2 = b"b" * 32

    with pytest.raises(OverflowError, match="int too big to convert"):
        # overflow
        Coin(H1, H2, 0x10000000000000000)  # type: ignore[arg-type]

    with pytest.raises(OverflowError, match="can't convert negative int to unsigned"):
        # overflow
        Coin(H1, H2, -1)  # type: ignore[arg-type]

    H1_short = b"a" * 31
    H1_long = b"a" * 33

    with pytest.raises(ValueError, match="could not convert slice to array"):
        # short hash
        Coin(H1_short, H2, uint64(1))

    with pytest.raises(ValueError, match="could not convert slice to array"):
        # long hash
        Coin(H1_long, H2, uint64(1))

    with pytest.raises(ValueError, match="could not convert slice to array"):
        # short hash
        Coin(H2, H1_short, uint64(1))

    with pytest.raises(ValueError, match="could not convert slice to array"):
        # long hash
        Coin(H2, H1_long, uint64(1))

    c = Coin(H1, H2, uint64(1000))
    assert c.parent_coin_info == H1
    assert c.puzzle_hash == H2
    assert c.amount == 1000



from __future__ import annotations

from typing import Optional

from klvm_tools.binutils import assemble

from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint16
from chik.wallet.nft_wallet.ownership_outer_puzzle import puzzle_for_ownership_layer
from chik.wallet.nft_wallet.transfer_program_puzzle import puzzle_for_transfer_program
from chik.wallet.outer_puzzles import construct_puzzle, get_inner_puzzle, get_inner_solution, match_puzzle, solve_puzzle
from chik.wallet.puzzle_drivers import PuzzleInfo, Solver
from chik.wallet.uncurried_puzzle import uncurry_puzzle


def test_ownership_outer_puzzle() -> None:
    ACS = Program.to(1)
    NIL = Program.to([])
    owner = bytes32.zeros
    # (mod (current_owner conditions solution)
    #     (list current_owner () conditions)
    # )
    transfer_program = assemble(
        """
        (c 2 (c () (c 5 ())))
        """
    )
    transfer_program_default: Program = puzzle_for_transfer_program(bytes32([1] * 32), bytes32([2] * 32), uint16(5000))
    ownership_puzzle: Program = puzzle_for_ownership_layer(owner, Program.to(transfer_program), ACS)
    ownership_puzzle_empty: Program = puzzle_for_ownership_layer(NIL, Program.to(transfer_program), ACS)
    ownership_puzzle_default: Program = puzzle_for_ownership_layer(owner, transfer_program_default, ACS)
    ownership_driver: Optional[PuzzleInfo] = match_puzzle(uncurry_puzzle(ownership_puzzle))
    ownership_driver_empty: Optional[PuzzleInfo] = match_puzzle(uncurry_puzzle(ownership_puzzle_empty))
    ownership_driver_default: Optional[PuzzleInfo] = match_puzzle(uncurry_puzzle(ownership_puzzle_default))
    transfer_program_driver: Optional[PuzzleInfo] = match_puzzle(uncurry_puzzle(transfer_program_default))

    assert ownership_driver is not None
    assert ownership_driver_empty is not None
    assert ownership_driver_default is not None
    assert transfer_program_driver is not None
    assert ownership_driver.type() == "ownership"
    assert ownership_driver["owner"] == owner
    assert ownership_driver_empty["owner"] == NIL
    assert ownership_driver["transfer_program"] == transfer_program
    assert ownership_driver_default["transfer_program"] == transfer_program_driver
    assert transfer_program_driver.type() == "royalty transfer program"
    assert transfer_program_driver["launcher_id"] == bytes32([1] * 32)
    assert transfer_program_driver["royalty_address"] == bytes32([2] * 32)
    assert transfer_program_driver["royalty_percentage"] == 5000
    assert construct_puzzle(ownership_driver, ACS) == ownership_puzzle
    assert construct_puzzle(ownership_driver_empty, ACS) == ownership_puzzle_empty
    assert construct_puzzle(ownership_driver_default, ACS) == ownership_puzzle_default
    assert get_inner_puzzle(ownership_driver, uncurry_puzzle(ownership_puzzle)) == ACS

    # Set up for solve
    inner_solution = Program.to(
        [
            [51, ACS.get_tree_hash(), 1],
            [-10],
        ]
    )
    solution: Program = solve_puzzle(
        ownership_driver,
        Solver({}),
        ACS,
        inner_solution,
    )
    ownership_puzzle.run(solution)
    assert get_inner_solution(ownership_driver, solution) == inner_solution
